// gcc -std=c99 -Wall -Wextra -pedantic -O3 coupled.c -lm -ljack

#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include <jack/jack.h>

static inline void f(double *out, const double x[2])
{
  double out0 = -exp(x[1] * x[1]) * x[0];
  double out1 = -exp(x[0] * x[0]) * x[1];
  out[0] = out0;
  out[1] = out1;
}

static inline double E(const double x[2], const double v[2])
{
  double w0 = sqrt(exp(x[1] * x[1]));
  double w1 = sqrt(exp(x[0] * x[0]));
  double k0 = 0.5 * v[0] * v[0];
  double k1 = 0.5 * v[1] * v[1];
  double u0 = 0.5 * w0 * x[0] * x[0];
  double u1 = 0.5 * w1 * x[1] * x[1];
  return k0 + k1 + u0 + u1;
}

jack_client_t *client;
jack_port_t *port[2];
double e;

static int processcb(jack_nframes_t nframes, void *arg)
{
  (void) arg;

  const double h = 0.01;
  static double x[2] = { 0, 1 };
  static double v[2] = { 1, 1 };
  static double a[2] = { 0, 0 };
  static int first = 1;
  if (first)
  {
    f(a, x);
    v[0] += 0.5 * h * a[0];
    v[1] += 0.5 * h * a[1];
    first = 0;
  }

  jack_default_audio_sample_t *out[2];
  out[0] = jack_port_get_buffer(port[0], nframes);
  out[1] = jack_port_get_buffer(port[1], nframes);

  for (jack_nframes_t i = 0; i < nframes; ++i)
  {
    out[0][i] = 0.5 * (x[0] += h * v[0]);
    out[1][i] = 0.5 * (x[1] += h * v[1]);
    f(a, x);
    v[0] += h * a[0];
    v[1] += h * a[1];
  }
  e = E(x, v);
  
  return 0;
}

#if 0
    // allows more accurate energy computation without fused v increments
    v[0] += 0.5 * h * a[0];
    v[1] += 0.5 * h * a[1];
    x[0] += h * v[0];
    x[1] += h * v[1];
    f(a, x);
    v[0] += 0.5 * h * a[0];
    v[1] += 0.5 * h * a[1];
    t += h;
    e = E(x, v);
#endif

int main(int argc, char **argv)
{
  (void) argc;
  (void) argv;
  if (! (client = jack_client_open("chaoscillators", JackNoStartServer, 0)))
  {
    fprintf(stderr, "jack server not running?\n");
    return 1;
  }
  jack_set_process_callback(client, processcb, 0);
  port[0] = jack_port_register(client, "output_1", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
  port[1] = jack_port_register(client, "output_2", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
  if (jack_activate(client))
  {
    fprintf (stderr, "cannot activate JACK client");
    return 1;
  }
  if (jack_connect(client, "chaoscillators:output_1", "system:playback_1"))
    fprintf(stderr, "cannot connect output port 1\n");
  if (jack_connect(client, "chaoscillators:output_2", "system:playback_2"))
    fprintf(stderr, "cannot connect output port 2\n");
  while (1)
  {
    sleep(1);
    fprintf(stderr, "\t%f\r", e);
  }
  return 0;
}
